/*
 * File:        mred.cc
 * Purpose:     MrEd main file
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_timer.h"
#include "wx_media.h"
#include <ctype.h>

#ifdef wx_motif
#include <X11/Xlib.h>
#include <X11/keysymdef.h>
#endif

#ifdef wx_msw
#include <dir.h>
#endif

#ifdef wx_x
#include <sys/types.h>
#include <sys/time.h>
#endif

#ifndef WXSCHEME
#define WXSCHEME 1
#endif

#ifndef REDIRECT_STDIO
#ifdef wx_msw
#define REDIRECT_STDIO 1
#else
#define REDIRECT_STDIO 0
#endif
#endif

#if WXSCHEME
#include "wxs/wxscheme.h"
#include "wxs/wxs_fram.h"

static Scheme_Env *global_env;

#ifdef wx_x
#define ADD_GC_DUMP 1
#else
#define ADD_GC_DUMP 0
#endif

#ifdef wx_motif
#define ERROR_HOOK
#endif

#if ADD_GC_DUMP
extern "C" { extern void GC_dump(); };
Scheme_Object *GCDump(int argc, Scheme_Object *argv[])
{
  GC_dump();
  return scheme_void;
}
#endif

#endif

class MrEdWindow : public wxFrame
{
  wxMediaCanvas *display;
  wxMediaEdit *media;

  int editOffset, fontOffset;

public:
  MrEdWindow(char *filename);
  ~MrEdWindow();

  void OnSize(int x, int y);
  void OnMenuCommand(int op);
};

#define MARGIN 2

#if 0
    asm(".globl mcount");
    asm("mcount:");
    asm("nop");
    asm("nop");
    asm("mov 0,%o0");
    asm("retl");
    asm("nop");
    asm("nop");
    asm("nop");
#endif

static int CheckForBreak(void);

class MrEdApp: public wxApp
{
#if WXSCHEME
  wxFrame *main;
#else
  MrEdWindow *main;
#endif
public:
  MrEdApp();
  wxFrame *OnInit(void);
};

#if WX_FORCE_APP_CREATION
MrEdApp *TheMrEdApp;
#else
MrEdApp TheMrEdApp;
#endif

MrEdApp::MrEdApp()
{
}

static void MrEdExit(int status)
{
  wxTheApp->ExitMainLoop();
}

#if WXSCHEME
static void MrEdIgnoreWarnings(char *msg, GC_word arg)
{
}

#if REDIRECT_STDIO
static FILE *mrerr = NULL;
#else
#define mrerr stderr
#endif

static void MrEdSchemeMessages(char *msg, ...)
{
  va_list args;

#if REDIRECT_STDIO
  if (!mrerr)
    mrerr = fopen("mrstderr.txt", "w");
  if (!mrerr)
    return;
#endif

  va_start(args, msg);
  vfprintf(mrerr, msg, args);
  va_end(args);
}

#if REDIRECT_STDIO

int stdin_getc(Scheme_Input_Port*)
{
  return EOF;
}

int stdin_char_ready(Scheme_Input_Port*)
{
  return TRUE;
}

static Scheme_Object *MrEdMakeStdIn(void)
{
  Scheme_Object *intype = scheme_make_port_type("stdin");

  return (Scheme_Object *)scheme_make_input_port(intype, NULL,
						 stdin_getc,
						 stdin_char_ready,
						 NULL, NULL);
}

void stdout_write(char *s, long l, Scheme_Output_Port*)
{
  static FILE *out = NULL;

  if (!out)
    out = fopen("mrstdout.txt", "w");
  
  if (out)
    fwrite(s, l, 1, out);
}

static Scheme_Object *MrEdMakeStdOut(void)
{
  Scheme_Object *outtype = scheme_make_port_type("stdout");

  return (Scheme_Object *)scheme_make_output_port(outtype, NULL,
						  stdout_write,
						  NULL);
}

void stderr_write(char *s, long l, Scheme_Output_Port*)
{
  if (!mrerr)
    mrerr = fopen("mrstderr.txt", "w");
  
  if (mrerr)
    fwrite(s, l, 1, mrerr);
}

static Scheme_Object *MrEdMakeStdErr(void)
{
  Scheme_Object *errtype = scheme_make_port_type("stderr");

  return (Scheme_Object *)scheme_make_output_port(errtype, NULL,
						  stderr_write,
						  NULL);
}
#endif

static void MrEdSleep(int secs, void *fds)
{
#ifdef wx_x
  if (fds) {
    int limit;
    fd_set rd, ex;
    
#ifdef USE_ULIMIT
    limit = ulimit(4, 0);
#else
    limit = getdtablesize();
#endif    
    
    memcpy(&rd, fds, sizeof(fd_set));
    memcpy(&ex, fds, sizeof(fd_set));
    select(limit, &rd, NULL, &ex, NULL);

    return;
  }
#endif
  wxSleep(secs);
}


#ifdef wx_x

#ifdef wx_motif
static XtInputId *scheme_cb_ids = NULL;
#endif
#ifdef wx_xview
static int *scheme_cb_ids = NULL;
#endif
static int num_cbs;

#ifdef wx_xview
Notify_value MrEdWakeUp(Notify_client client, int id)
#endif
#ifdef wx_motif
Bool MrEdWakeUp(XtPointer client, int *fid, XtInputId * id)
#endif
{
  int i;

  if (scheme_cb_ids) {
#ifdef wx_motif
    if (scheme_cb_ids)
      for (i = 0; i < num_cbs; i++)
	XtRemoveInput(scheme_cb_ids[i]);
#endif
#ifdef wx_xview
    if (scheme_cb_ids)
      for (i = 0; i < num_cbs; i++)
	notify_set_input_func((Notify_client)NULL, (Notify_func)NULL, 
			      scheme_cb_ids[i]);
#endif

    scheme_cb_ids = NULL;
    
    scheme_wake_up();
  }

#ifdef wx_motif
  return FALSE;
#else
  return (Notify_value)FALSE;
#endif
}

static void MrEdNeedWakeup(void *fds)
{
  int limit, count, i, p;

#ifdef USE_ULIMIT
    limit = ulimit(4, 0);
#else
    limit = getdtablesize();
#endif    
  count = 0;
  for (i = 0; i < limit; i++)
    if (FD_ISSET(i, (fd_set *)fds))
      count++;

  if (!count)
    return;

#ifdef wx_motif
  if (scheme_cb_ids)
    for (i = 0; i < num_cbs; i++)
      XtRemoveInput(scheme_cb_ids[i]);

  num_cbs = 2 * count;
  scheme_cb_ids = new XtInputId[num_cbs];
#endif
#ifdef wx_xview
  if (scheme_cb_ids)
    for (i = 0; i < num_cbs; i++)
      notify_set_input_func((Notify_client)NULL, (Notify_func)NULL, 
			    scheme_cb_ids[i]);

  num_cbs = count;
  scheme_cb_ids = new int[num_cbs];
#endif


  p = 0;
  for (i = 0; i < limit; i++)
    if (FD_ISSET(i, (fd_set *)fds)) {
#ifdef wx_xview
      notify_set_input_func((Notify_client)NULL, (Notify_func)MrEdWakeUp, i);
      scheme_cb_ids[p] = i;
#endif
#ifdef wx_motif
      scheme_cb_ids[p] = XtAppAddInput(wxTheApp->appContext, i, 
				       (XtPointer *)XtInputReadMask, 
				       (XtInputCallbackProc)MrEdWakeUp, NULL);
      scheme_cb_ids[p+count] = XtAppAddInput(wxTheApp->appContext, i, 
					     (XtPointer *)XtInputExceptMask, 
					     (XtInputCallbackProc)MrEdWakeUp, 
					     NULL);
#endif
      p++;
    }
}
#endif

class MrEdThreadTimer : public wxTimer
{
 public:
  void Notify(void);
};

static int threads_go;
static MrEdThreadTimer *theThreadTimer;

#define THREAD_WAIT_TIME 100

void MrEdThreadTimer::Notify()
{
  scheme_check_threads();

  if (threads_go)
    Start(THREAD_WAIT_TIME, TRUE);
}

static void MrEdInstallThreadTimer(int on)
{
  if (!theThreadTimer)
    theThreadTimer = new MrEdThreadTimer;

  if (on)
    theThreadTimer->Start(THREAD_WAIT_TIME, TRUE);
  else
    theThreadTimer->Stop();

  threads_go = on;
}

#endif

#ifdef ERROR_HOOK
XtErrorHandler old_handler;

void ErrorHook(String msg)
{
  old_handler(msg);
}
#endif

wxFrame *MrEdApp::OnInit(void)
{
  wxInitMedia();

#ifdef ERROR_HOOK
  old_handler = XtAppSetErrorHandler(appContext, ErrorHook);
#endif

#if WXSCHEME
#define DIR_BUF_SIZE 1000
  char save_buf[DIR_BUF_SIZE], filename[1000];
  Bool ok;
  static wxFrame *real_main; /* Keeps it from being GC'd */

#if REDIRECT_STDIO
  scheme_make_stdin = MrEdMakeStdIn;
  scheme_make_stdout = MrEdMakeStdOut;
  scheme_make_stderr = MrEdMakeStdErr;
#endif

  GC_set_warn_proc(MrEdIgnoreWarnings);
  scheme_console_printf = MrEdSchemeMessages;
  scheme_sleep = MrEdSleep;
  scheme_notify_multithread = MrEdInstallThreadTimer;
#ifdef wx_x
  scheme_wakeup_on_input = MrEdNeedWakeup;
#endif

  main = real_main = new wxFrame(NULL, "MrEd");
  wxPanel *p = new wxPanel(main);
  (void)(new wxButton(p, (wxFunction)wxExit, "Quit"));
  p->Fit();
  main->Fit();

  scheme_case_sensitive = 1;
  scheme_constant_builtins = 1;

  global_env = wxsScheme_setup();

  scheme_exit = MrEdExit;

#ifdef wx_x
  scheme_add_global_constant("mred:platform", 
			     scheme_intern_symbol("unix"), global_env);
#ifdef wx_motif
  scheme_add_global_constant("mred:window-system", 
			     scheme_intern_symbol("motif"), global_env);
#else
  scheme_add_global_constant("mred:window-system", 
			     scheme_intern_symbol("xview"), global_env);
#endif  
#else
  scheme_add_global_constant("mred:platform", 
			     scheme_intern_symbol("windows"), global_env);
  scheme_add_global_constant("mred:window-system", 
			     scheme_intern_symbol("windows"), global_env);
#endif  

#if ADD_GC_DUMP
  scheme_add_global("dump-memory-statistics", 
		    scheme_make_prim(GCDump), global_env);
#endif

  wxPathList *paths = new wxPathList();
  char *file;

#ifdef wx_x
  /* Skip command name */
  argv++;
  --argc;
#endif

  if (argc >= 2 && !strcmp(argv[0], "-s")) {
    file = argv[1];
    argc -=2;
    argv += 2;
  } else {
    paths->Add(".");
    paths->AddEnvList("MREDSYS");
    paths->Add("./system/");
    file = paths->FindValidPath("mrsystem.ss");
    if (!file) {
      wxMessageBox("Can't find \"mrsystem.ss\".", "Error");
      goto giveup;
    } else
      file = copystring(file);
    delete paths;
  }

  wxGetWorkingDirectory(save_buf, DIR_BUF_SIZE);
  chdir(PathOnly(file));
  strcpy(filename, FileNameFromPath(file));
  ok = scheme_load(filename);
  chdir(save_buf);
  if (!ok){
    wxMessageBox("Error loading system.", "Error");
    goto giveup;
  }

  if (argc < 1 || (strcmp(argv[0],"-q") 
		   && strcmp(argv[0], "--no-init-file"))) {
    Scheme_Object *setup;

    setup = scheme_lookup_global(scheme_intern_symbol("mred:user-setup"), 
				 global_env);

    if (!setup) {
      wxMessageBox("Can't find mred:user-setup function.", "Error");
      goto giveup;
    }

    if (setjmp(scheme_error_buf)) {
      wxMessageBox("Error in user setup.", "Error");
      goto giveup;
    } else
      scheme_apply(setup, 0, NULL);
  } else {
    --argc;
    argv++;
  }

  Scheme_Object *startup;

  startup = scheme_lookup_global(scheme_intern_symbol("mred:startup"), 
				 global_env);

  if (!startup) {
    wxMessageBox("Can't find mred:startup function.", "Error");
    goto giveup;
  }

  if (setjmp(scheme_error_buf)) {
    wxMessageBox("Error starting up.", "Error");
    goto giveup;
  } else {
    Scheme_Object **p;
    int i;

    p = (Scheme_Object **)scheme_malloc(argc * sizeof(Scheme_Object *));
    for (i = 0; i < argc; i++)
      p[i] = scheme_make_string(argv[i]);

    startup = scheme_apply(startup, argc, p);

    if (objscheme_istype_wxFrame(startup, NULL, 0))
      main = objscheme_unbundle_wxFrame(startup, NULL, 0);
#ifdef wx_motif
    else
      main->Show(TRUE);
#endif
  }

  scheme_check_for_break = CheckForBreak;

  goto done;
 giveup:
  main = NULL;
 done:
#else
  main = NULL;
#endif

  if (!main)
    main = new MrEdWindow(NULL);

  return main;
}

#if WX_FORCE_APP_CREATION
extern "C" { int wxEntry(int argc, char *argv[]); };

/* Solves C++ problem of intializing TheApp for some compilers */
int main(int argc, char *argv[])
{
  TheMrEdApp = new MrEdApp;
  return wxEntry(argc, argv);
}
#endif

/****************************************************************/

/* Break-checking is platform-dependent. Oh, well... */

#if defined(wx_motif) || defined(wx_xview)
static Bool CheckPred(Display *d, XEvent *e, char *args)
{
  if (e->type == KeyPress) {
    XKeyEvent *key;

    key = (XKeyEvent *)e;
    return (key->state & ControlMask) && (key->state & ShiftMask) 
      && (key->keycode == *(KeyCode *)args);
  } else
    return FALSE;
}
#endif

static int CheckForBreak(void)
{
#ifdef wx_x
  XEvent e;
  static Display *d = NULL;
  static KeyCode code;

  if (!d) {
#ifdef wx_motif
    d = XtDisplay(wxTheApp->topLevel);
#else
    d = (Display*)xv_get((Frame)(wxTheApp->wx_frame->GetHandle()), XV_DISPLAY);
#endif
    code = XKeysymToKeycode(d, 'c');
  }

  XFlush(d);
  if (XCheckIfEvent(d, &e, CheckPred, (char *)&code))
    return TRUE;
#endif
#ifdef wx_msw
  MSG msg;
  MSG *msgs = NULL;
  int msgs_size = 0, msgs_count = 0, i;
  Bool ctlDown = FALSE, shiftDown = FALSE;

  if (!::PeekMessage(&msg, NULL, WM_KEYDOWN, WM_KEYDOWN, PM_NOREMOVE))
    return FALSE;

  // To search all the messages, we have to Peek them all and then
  // Post them back. This will mess up the timing and locations of
  // these events.
  //
  // Is there a better way to do this???

  while (::PeekMessage(&msg, NULL, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE)) {
    if (msg.message == WM_KEYDOWN) {
      if (msg.wParam == VK_CONTROL)
	ctlDown = TRUE;
      else if (msg.wParam == VK_SHIFT)
	shiftDown = TRUE;
      else if (shiftDown && ctlDown && (msg.wParam == 'C')) {
	delete[] msgs;
	return TRUE;
      }
    } else if (msg.message == WM_KEYUP) {
      if (msg.wParam == VK_CONTROL)
	ctlDown = FALSE;
      else if (msg.wParam == VK_SHIFT)
	shiftDown = FALSE;
    }
    if (msgs_count == msgs_size) {
      MSG *save = msgs;

      msgs_size = (msgs_size ? 2 * msgs_size : 10);
      msgs = new MSG[msgs_size];

      memcpy(msgs, save, msgs_count * sizeof(MSG));

      delete[] save;
    }
    memcpy(msgs + (msgs_count++), &msg, sizeof(MSG));
  }

  for (i = 0; i < msgs_count; i++)
    ::PostMessage(msgs[i].hwnd, msgs[i].message, 
		  msgs[i].wParam, msgs[i].lParam);

  delete[] msgs;
#endif
  return FALSE;
}

/****************************************************************/

enum {
  MrEdFILE_OPEN = 1,
  MrEdFILE_SAVE,
  MrEdFILE_SAVE_AS,
  MrEdFILE_SAVE_AS_TEXT,
  MrEdFILE_SAVE_AS_FORMATTED,
  MrEdFILE_RELOAD,
  MrEdFILE_QUIT
  };

MrEdWindow::MrEdWindow(char *filename)
: wxFrame(NULL, "MrEd", -1, -1, 600, 800, wxDEFAULT_FRAME + wxSDI, "frame")
{
  wxMenuBar *menubar;
  wxMenu *file, *edit;

  display = new wxMediaCanvas(this);

  media = new wxMediaEdit();
  display->SetMedia(media);

  menubar = new wxMenuBar();

  file = new wxMenu();
  file->Append(MrEdFILE_OPEN, "Open...");
  file->Append(MrEdFILE_SAVE, "Save");
  file->Append(MrEdFILE_SAVE_AS, "Save As...");
  file->Append(MrEdFILE_SAVE_AS_TEXT, "Save As Text...");
  file->Append(MrEdFILE_SAVE_AS_FORMATTED, "Save As Formatted...");
  file->Append(MrEdFILE_RELOAD, "Reload");  
  file->AppendSeparator();
  file->Append(MrEdFILE_QUIT, "Close");
    
  editOffset = MrEdFILE_QUIT + 1;

  edit = new wxMenu();
  fontOffset = media->AppendEditItems(edit, editOffset);
  edit->AppendSeparator();
  media->AppendFontItems(edit, fontOffset);
  
  menubar->Append(file, "File");
  menubar->Append(edit, "Edit");

  SetMenuBar(menubar);

  Show(TRUE);

  if (filename)
    media->LoadFile(filename);
}

MrEdWindow::~MrEdWindow()
{
  delete display;
}

void MrEdWindow::OnMenuCommand(int op)
{
  if (!display)
    return;

  if (op >= fontOffset)
    media->DoFont(op - fontOffset);
  else if (op >= editOffset)
    media->DoEdit(op - editOffset);
  else {
    switch(op) {
    case MrEdFILE_OPEN:      
      media->LoadFile("");
      break;
    case MrEdFILE_SAVE:
      media->SaveFile();
      break;
    case MrEdFILE_SAVE_AS:
      media->SaveFile("");
      break;
    case MrEdFILE_SAVE_AS_TEXT:
      media->SaveFile("", wxMEDIA_FF_TEXT);
      break;
    case MrEdFILE_SAVE_AS_FORMATTED:
      media->SaveFile("", wxMEDIA_FF_STD);
      break;
    case MrEdFILE_RELOAD:
      media->LoadFile();
      break;
    case MrEdFILE_QUIT:
      delete this;
      break;
    }
  }
}

void MrEdWindow::OnSize(int x, int y)
{
  int top_h;

  GetClientSize(&x, &y);

  top_h = 0;

  if (display)
    display->SetSize(MARGIN, MARGIN + top_h, 
		     x - 2 * MARGIN, y - (2 * MARGIN + top_h));
}
