/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

/* MzScheme function prototypes */
/* No macros should go in this file; it is used both to
   prototype functions, and as a parsing source for
   declaring scheme_extension_table */

/* The scheme_extension_table parser is touchy: don't leave a space
   between a function name and it's opening parameter parenthesis. */

/* After this START tag, all comments should start & end on same line */

/* START */

/* Call/cc utilities */
void scheme_init_jmpup_buf(Scheme_Jumpup_Buf *b);
int scheme_setjmpup_relative(Scheme_Jumpup_Buf *b, void *start, 
			     Scheme_Jumpup_Buf *cont);
void scheme_longjmpup(Scheme_Jumpup_Buf *b);

/* Processes */
extern Scheme_Process **scheme_current_process_ptr;

Scheme_Object *scheme_spawn(Scheme_Object *expr);
Scheme_Object *scheme_thread(Scheme_Object *thunk);

void scheme_process_block(int sleep_time);
void scheme_swap_process(Scheme_Process *process);

void scheme_check_threads(void);
void scheme_wake_up(void);

/* error handling */
void scheme_signal_error(char *msg, ...);
void scheme_warning(char *msg, ...);

void scheme_wrong_count(const char *name, int minc, int maxc, int argc,
			Scheme_Object **argv);
void scheme_wrong_type(const char *name, const char *expected, 
		       int which, int argc,
		       Scheme_Object **argv);


Scheme_Object *scheme_dynamic_wind(void (*pre)(void *),
				   Scheme_Object *(*act)(void *),
				   void (*post)(void *), 
				   Scheme_Object *(*jmp_handler)(void *),
				   void *data);

/* Types */
Scheme_Type scheme_make_type(const char *name);

/* Type readers & writers for compiled code data */
void scheme_install_type_reader(Scheme_Type type, Scheme_Type_Reader f);
void scheme_install_type_writer(Scheme_Type type, Scheme_Type_Writer f);

/* Constants */
extern Scheme_Object *scheme_eof;
extern Scheme_Object *scheme_null;
extern Scheme_Object *scheme_true;
extern Scheme_Object *scheme_false;
extern Scheme_Object *scheme_void;

/* Basics */
Scheme_Object *scheme_eval(Scheme_Object *obj, Scheme_Env *env);
Scheme_Object *scheme_eval_compiled(Scheme_Object *obj, Scheme_Env *env);
Scheme_Object *scheme_apply(Scheme_Object *f, int c, Scheme_Object **args);
Scheme_Object *scheme_apply_to_list(Scheme_Object *f, Scheme_Object *argss);
Scheme_Object *scheme_eval_string(const char *str, Scheme_Env *env);

/* Tail calls - only use these when you're writing new functions/syntax */
Scheme_Object *scheme_tail_apply(Scheme_Object *f, int n, Scheme_Object **arg);
Scheme_Object *scheme_tail_apply_to_list(Scheme_Object *f, Scheme_Object *l);

Scheme_Object *scheme_tail_eval_expr(Scheme_Object *obj, Scheme_Env *env);

/* Allocation */
#ifndef SCHEME_NO_GC
void *GC_malloc(size_t size_in_bytes);
void *GC_malloc_atomic(size_t size_in_bytes);
void *GC_malloc_stubborn(size_t size_in_bytes);
void *GC_malloc_uncollectable(size_t size_in_bytes);
#endif

void *scheme_malloc_eternal(size_t n);
void scheme_end_stubborn_change(void *p);

void *scheme_calloc(size_t num, size_t size);

char *scheme_strdup(const char *str);
char *scheme_strdup_eternal(const char *str);

/* Hash table */
Scheme_Hash_Table *scheme_hash_table(int size, int type, 
				      int w_const, int forever);
void scheme_add_to_table(Scheme_Hash_Table *table, const char *key, void *val, int);
void scheme_change_in_table(Scheme_Hash_Table *table, const char *key, void *new_val);
void *scheme_lookup_in_table(Scheme_Hash_Table *table, const char *key);
Scheme_Bucket *scheme_bucket_from_table(Scheme_Hash_Table *table, const char *key);

/* Constructors */
Scheme_Object *scheme_make_prim(Scheme_Prim *prim);
Scheme_Object *scheme_make_noneternal_prim(Scheme_Prim *prim);
Scheme_Object *scheme_make_closed_prim(Scheme_Closed_Prim *prim, void *data);
Scheme_Object *scheme_make_prim_w_arity(Scheme_Prim *prim, short mina, 
					short maxa);
Scheme_Object *scheme_make_noneternal_prim_w_arity(Scheme_Prim *prim, 
						   short mina, short maxa);
Scheme_Object *scheme_make_closed_prim_w_arity(Scheme_Closed_Prim *prim, 
					       void *data, short mina, 
					       short maxa);
Scheme_Object *scheme_make_closure(Scheme_Env *env, Scheme_Object *code);
Scheme_Object *scheme_make_pair(Scheme_Object *car, Scheme_Object *cdr);
Scheme_Object *scheme_make_string(const char *chars);
Scheme_Object *scheme_make_sized_string(char *chars, long len, int copy);
Scheme_Object *scheme_make_string_without_copying(char *chars);
Scheme_Object *scheme_alloc_string(int size, char fill);
Scheme_Object *scheme_append_string(Scheme_Object *, Scheme_Object *);
Scheme_Object *scheme_make_vector(int size, Scheme_Object *fill);
Scheme_Object *scheme_make_integer_value(long i);
Scheme_Object *scheme_make_double(double d);
Scheme_Object *scheme_make_char(char ch);
Scheme_Object *scheme_make_compiled_syntax(Scheme_Syntax *syntax,
					   Scheme_Syntax_Expander *exp);
Scheme_Object *scheme_make_promise(Scheme_Object *expr, Scheme_Env *env);
Scheme_Object *scheme_make_trigger();

/* Bignums */
Scheme_Object *scheme_make_bignum(int v);
double scheme_bignum_to_float(Scheme_Object *n);
Scheme_Object *scheme_bignum_from_float(double d);
char *scheme_bignum_to_string(Scheme_Object *n, int radix);
Scheme_Object *scheme_read_bignum(const char *str, int radix);
Scheme_Object *scheme_bignum_normalize(Scheme_Object *n);

/* Macro and syntax expansion */
Scheme_Object *scheme_expand(Scheme_Object *form, Scheme_Env *env);
Scheme_Object *scheme_expand_expr(Scheme_Object *form, Scheme_Env *env,
				  int depth);
Scheme_Object *scheme_expand_list(Scheme_Object *form, Scheme_Env *env,
				  int depth);
Scheme_Object *scheme_expand_block(Scheme_Object *form, Scheme_Env *env,
				   int depth);
Scheme_Object *scheme_expand_simple(Scheme_Object *form, Scheme_Env *env,
				    int depth);

/* Compilation */
Scheme_Object *scheme_compile(Scheme_Object *form, Scheme_Env *env);
Scheme_Object *scheme_compile_expr(Scheme_Object *form, Scheme_Env *env,
				   Scheme_Compile_Info *rec);
Scheme_Object *scheme_compile_simple(Scheme_Object *form, Scheme_Env *env,
				     Scheme_Compile_Info *rec);
Scheme_Object *scheme_compile_sequence(Scheme_Object *forms, Scheme_Env *env,
			      Scheme_Compile_Info *rec);
Scheme_Object *scheme_compile_block(Scheme_Object *forms, Scheme_Env *env,
			      Scheme_Compile_Info *rec);
Scheme_Object *scheme_compile_list(Scheme_Object *form, Scheme_Env *env,
			      Scheme_Compile_Info *rec);

void scheme_init_compile_recs(Scheme_Compile_Info *src, 
			      Scheme_Compile_Info *dest, int n);
void scheme_merge_compile_recs(Scheme_Compile_Info *src, 
			       Scheme_Compile_Info *dest, int n);

Scheme_Object *scheme_make_syntax_compilation(Scheme_Syntax_Compilation *prim,
					      Scheme_Object *data);
Scheme_Object *scheme_make_closure_compilation(Scheme_Env *env,
					       Scheme_Object *uncompiled_code,
					       Scheme_Compile_Info *rec);
Scheme_Object *scheme_make_compiled_closure(Scheme_Env *env, 
					    Scheme_Object *compiled_code);
Scheme_Object *scheme_make_promise_compiled(Scheme_Object *expr, 
					    Scheme_Env *env);
Scheme_Object *scheme_make_branch_compilation(Scheme_Object *test,
					      Scheme_Object *tbranch,
					      Scheme_Object *fbranch);
Scheme_Object *scheme_make_sequence_compilation(Scheme_Object *compiled_list);

void scheme_register_syntax_executer(const char *name, Scheme_Syntax_Compilation *f);

/* Ports */
Scheme_Object *scheme_read(Scheme_Object *port);
void scheme_write(Scheme_Object *obj, Scheme_Object *port);
void scheme_display(Scheme_Object *obj, Scheme_Object *port);
void scheme_write_string(const char *str, long len, Scheme_Object *port);
char *scheme_write_to_string(Scheme_Object *obj, long *len);
char *scheme_display_to_string(Scheme_Object *obj, long *len);
void scheme_debug_print(Scheme_Object *obj);
void scheme_flush_output(Scheme_Object *port);

int scheme_getc(Scheme_Object *port);
void scheme_ungetc(int ch, Scheme_Object *port);
int scheme_char_ready(Scheme_Object *port);
void scheme_need_wakeup(Scheme_Object *port, void *fds);
long scheme_tell(Scheme_Object *port);
long scheme_tell_line(Scheme_Object *port);
void scheme_close_input_port(Scheme_Object *port);
void scheme_close_output_port(Scheme_Object *port);

Scheme_Object *scheme_make_port_type(const char *name);
Scheme_Input_Port *scheme_make_input_port(Scheme_Object *subtype, void *data,
					  int (*getc_fun)(Scheme_Input_Port*),
					  int (*char_ready_fun)
					  (Scheme_Input_Port*),
					  void (*close_fun)
					  (Scheme_Input_Port*),
					  void (*need_wakeup_fun)
					  (Scheme_Input_Port*, void *));
Scheme_Output_Port *scheme_make_output_port(Scheme_Object *subtype,
					    void *data,
					    void (*write_string_fun)
					    (char*,long, Scheme_Output_Port*),
					    void (*close_fun)
					    (Scheme_Output_Port*));

Scheme_Object *scheme_make_file_input_port(FILE *fp);
Scheme_Object *scheme_make_named_file_input_port(FILE *fp, const char *filename);
Scheme_Object *scheme_make_file_output_port(FILE *fp);

Scheme_Object *scheme_make_string_input_port(const char *str);
Scheme_Object *scheme_make_string_output_port();
char *scheme_get_string_output(Scheme_Object *);

void scheme_pipe(Scheme_Object **write, Scheme_Object **read);

int scheme_file_exists(char *filename);

/* environment */
void scheme_add_global(const char *name, Scheme_Object *val, Scheme_Env *env);
void scheme_add_global_constant(const char *name, Scheme_Object *v, Scheme_Env *env);
void scheme_add_global_keyword(const char *name, Scheme_Object *v, Scheme_Env *env);
void scheme_remove_global(const char *name, Scheme_Env *env);

void scheme_add_global_symbol(Scheme_Object *name, Scheme_Object *val, 
			      Scheme_Env *env);
void scheme_remove_global_symbol(Scheme_Object *name, Scheme_Env *env);

void scheme_check_identifier(const char *form, Scheme_Object *id, 
			     const char *where,
			     Scheme_Env *env);

void scheme_constant(Scheme_Object *sym, Scheme_Env *env);

Scheme_Env *scheme_new_compilation_frame(int num_bindings, int flags);
void scheme_add_compilation_binding(int index, Scheme_Object *val, 
				    Scheme_Env *frame);
Scheme_Env *scheme_add_compilation_frame(Scheme_Object *vals, 
					 Scheme_Env *env, int flags);

Scheme_Env *scheme_lengthen_compilation_frame(int count, Scheme_Env *frame);
void scheme_frame_backinfo(Scheme_Env *orig, Scheme_Env *extended);

Scheme_Env *scheme_new_special_frame(int num_bindings, int flags);
void scheme_add_binding(int index, Scheme_Object *val, Scheme_Env *frame);
Scheme_Env *scheme_add_frame(Scheme_Object *vals, Scheme_Env *env, int flags);

Scheme_Object *scheme_annotate_frame_id(Scheme_Env *env, Scheme_Object *name);

Scheme_Object *scheme_make_envunbox(Scheme_Object *value);

Scheme_Object *scheme_lookup_value(Scheme_Object *symbol, Scheme_Env *env);
Scheme_Object *scheme_lookup_global(Scheme_Object *symbol, Scheme_Env *env);
Scheme_Bucket *scheme_global_bucket(Scheme_Object *symbol, Scheme_Env *env);
Scheme_Object *scheme_static_distance(Scheme_Object *symbol, Scheme_Env *env,
				      int elim_const_lookup);

void scheme_set_value(Scheme_Object *sd, Scheme_Object *val, Scheme_Env *env);

/* Symbols */
Scheme_Object *scheme_intern_symbol(const char *name);
Scheme_Object *scheme_make_symbol(const char *name); /* Make uninterned */
Scheme_Object *scheme_intern_exact_symbol(const char *name); /* Exact case */

/* Type Symbols */
Scheme_Object *scheme_intern_type_symbol(Scheme_Object *sym);
Scheme_Object *scheme_make_type_symbol(Scheme_Object *sym);/*Make uninterned */

/* Objects */
Scheme_Object *scheme_make_class(const char *name, Scheme_Object *sup, 
				 Scheme_Method_Prim *init);
void scheme_add_method(Scheme_Object *cl, const char *name, 
		       Scheme_Method_Prim *f);

Scheme_Object *scheme_make_object(Scheme_Object *sclass, 
				  int argc, Scheme_Object **argv);
Scheme_Object *scheme_make_uninited_object(Scheme_Object *sclass);

Scheme_Object *scheme_find_ivar(Scheme_Object *obj, Scheme_Object *sclass, 
				Scheme_Object *sym, int force);

int scheme_is_subclass(Scheme_Object *sub, Scheme_Object *parent);
int scheme_is_a(Scheme_Object *obj, Scheme_Object *sclass);
char *scheme_get_class_name(Scheme_Object *cl);

Scheme_Object *scheme_apply_generic_data(Scheme_Object *gdata, 
					 Scheme_Object *sobj, int force);
Scheme_Object *scheme_get_generic_data(Scheme_Object *cl, 
				       Scheme_Object *name);

/* Misc */
int scheme_eq(Scheme_Object *obj1, Scheme_Object *obj2);
int scheme_eqv(Scheme_Object *obj1, Scheme_Object *obj2);
int scheme_equal(Scheme_Object *obj1, Scheme_Object *obj2);

int scheme_list_length(Scheme_Object *list);

Scheme_Object *scheme_alloc_list(int size);
Scheme_Object *scheme_map_1(Scheme_Object *(*f)(Scheme_Object*),
			    Scheme_Object *l);

Scheme_Object *scheme_car(Scheme_Object *pair);
Scheme_Object *scheme_cdr(Scheme_Object *pair);
Scheme_Object *scheme_cadr(Scheme_Object *pair);
Scheme_Object *scheme_caddr(Scheme_Object *pair);

Scheme_Object *scheme_vector_to_list(Scheme_Object *vec);
Scheme_Object *scheme_list_to_vector(Scheme_Object *list);

Scheme_Object *scheme_append(Scheme_Object *lstx, Scheme_Object *lsty);

Scheme_Object *scheme_box(Scheme_Object *v);
Scheme_Object *scheme_unbox(Scheme_Object *obj);
void scheme_set_box(Scheme_Object *b, Scheme_Object *v);

int scheme_load(const char *file);
void scheme_load_extension(const char *filename);
void scheme_register_extension_global(void *ptr, long size);

void scheme_rep(void);
char *scheme_banner(void);
char *scheme_version(void);
