/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"

static Scheme_Object *make_trigger(int n, Scheme_Object **p);
static Scheme_Object *hit_trigger(int n, Scheme_Object **p);
static Scheme_Object *hit_trigger_p(int n, Scheme_Object **p);
static Scheme_Object *block_trigger(int n, Scheme_Object **p);
static Scheme_Object *trigger_callback(int n, Scheme_Object **p);
static Scheme_Object *trigger_input_port(int n, Scheme_Object **p);

void scheme_init_trigger(Scheme_Env *env)
{
  scheme_add_global_constant("make-trigger", 
			     scheme_make_prim_w_arity(make_trigger, 0, 0), 
			     env);
  scheme_add_global_constant("trigger-hit", 
			     scheme_make_prim_w_arity(hit_trigger, 1, 1), 
			     env);
  scheme_add_global_constant("trigger-hit?", 
			     scheme_make_prim_w_arity(hit_trigger_p, 1, 1), 
			     env);
  scheme_add_global_constant("trigger-block", 
			     scheme_make_prim_w_arity(block_trigger, 1, 2), 
			     env);
  scheme_add_global_constant("trigger-callback", 
			     scheme_make_prim_w_arity(trigger_callback, 2, 2), 
			     env);
  scheme_add_global_constant("input-port-trigger", 
			     scheme_make_prim_w_arity(trigger_input_port, 
						      2, 2), 
			     env);
}

Scheme_Object *scheme_make_trigger(void)
{
  Scheme_Trigger *trigger;

  trigger = (Scheme_Trigger *)scheme_malloc_atomic(sizeof(Scheme_Trigger));
  trigger->type = scheme_trigger_type;
  trigger->value = 0;

  return (Scheme_Object *)trigger;
}

static Scheme_Object *make_trigger(int n, Scheme_Object **p)
{
  if (n)
    scheme_wrong_count("make-trigger", 0, 0, n, p);

  return scheme_make_trigger();
}

static Scheme_Object *hit_trigger(int n, Scheme_Object **p)
{
  if (n != 1)
    scheme_wrong_count("trigger-hit", 1, 1, n, p);
  if (!SCHEME_TRIGGERP(p[0]))
    scheme_wrong_type("trigger-hit", "trigger", 0, n, p);

  ((Scheme_Trigger *)p[0])->value++;

  return scheme_void;
}

static Scheme_Object *hit_trigger_p(int n, Scheme_Object **p)
{
  if (n != 1)
    scheme_wrong_count("trigger-hit?", 1, 1, n, p);
  if (!SCHEME_TRIGGERP(p[0]))
    scheme_wrong_type("trigger-hit?", "trigger", 0, n, p);

  return (((Scheme_Trigger *)p[0])->value ? scheme_true : scheme_false);
}

static Scheme_Object *block_trigger(int n, Scheme_Object **p)
{
  int start;
  Scheme_Trigger *trigger;

  if (n != 1 && n != 2)
    scheme_wrong_count("trigger-block", 1, 2, n, p);
  if (!SCHEME_TRIGGERP(p[0]))
    scheme_wrong_type("trigger-block", "trigger", 0, n, p);

  trigger = (Scheme_Trigger *)p[0];

  if (n == 2 && NOT_SAME_OBJ(p[1], scheme_false))
    start = 0;
  else
    start = trigger->value;

  scheme_current_process->block_descriptor = TRIGGER_BLOCKED;
  scheme_current_process->blocker = (Scheme_Object *)trigger;

  while (trigger->value == start) {
    scheme_process_block(0);
  }

  scheme_current_process->block_descriptor = NOT_BLOCKED;
  scheme_current_process->blocker = NULL;
  scheme_current_process->ran_some = 1;

  return scheme_void;
}

static Scheme_Object *trigger_callback(int n, Scheme_Object **p)
{
  Scheme_Trigger_Callback *cb;

  if (n != 2)
    scheme_wrong_count("trigger-callback", 2, 2, n, p);
  if (!SCHEME_TRIGGERP(p[0]))
    scheme_wrong_type("trigger-callback", "trigger", 0, n, p);
  if (!SCHEME_PROCP(p[1]))
    scheme_wrong_type("trigger-callback", "procedure", 1, n, p);

  cb = (Scheme_Trigger_Callback*)scheme_malloc(sizeof(Scheme_Trigger_Callback));
  cb->trigger = (Scheme_Trigger *)p[0];
  cb->start = cb->trigger->value;
  cb->callback = p[1];
  
  scheme_add_trigger_callback(cb);

  return scheme_void;
}

static Scheme_Object *trigger_input_port(int n, Scheme_Object **p)
{
  Scheme_Trigger_From_Port *cb;

  if (n != 2)
    scheme_wrong_count("input-port-trigger", 2, 2, n, p);
  if (!SCHEME_INPORTP(p[0]))
    scheme_wrong_type("input-port-trigger", "input port", 0, n, p);
  if (!SCHEME_TRIGGERP(p[1]))
    scheme_wrong_type("input-port-trigger", "trigger", 1, n, p);

  cb = (Scheme_Trigger_From_Port*)scheme_malloc(sizeof(Scheme_Trigger_From_Port));
  cb->trigger = (Scheme_Trigger *)p[1];
  cb->port = p[0];
  
  scheme_add_trigger_from_port(cb);

  return scheme_void;
}

