(define mred:std-buffer-group%
  (let ([untitled-number 0])
    (let-struct buffer (name type edit modified-ok)
       (make-class ()
		   (private
		    [buffers '()])
		   (public
		    [find-buffer
		     (opt-lambda (type pred [otherwise (lambda () #f)])
				 (catch exit
					 (for-each
					  (lambda (b)
					    (if (and (eq? (buffer-type b) type)
						     (pred (buffer-name b)
							   (buffer-edit b)))
						(exit (buffer-edit b))))
					  buffers)
					 (otherwise)))]
		    [add-buffer
		     (lambda (type name edit)
		       (set! buffers (cons (make-buffer name type edit #f) 
					   buffers)))]
		    [kill-buffer
		     (lambda (edit)
		       (if (not (null? buffers))
			   (if (eq? edit (buffer-edit (car buffers)))
			       (set! buffers (cdr buffers))
			       (let loop ([buffers buffers])
				 (if (not (null? (cdr buffers)))
				     (if (eq? edit (buffer-edit (cadr buffers)))
					 (set-cdr! buffers (cddr buffers))
					 (loop (cdr buffers))))))))]

		    [reset-buffer-state
		     (lambda (edit)
		       (for-each
			(lambda (buffer)
			  (if (eq? (buffer-edit buffer) edit)
			      (set-buffer-modified-ok! buffer #f)))
			buffers))]

		    [check-buffers
		     (lambda ()
		       (let loop ([buffers buffers])
			 (if (not (null? buffers))
			     (let* ([buffer (car buffers)]
				    [type (buffer-type buffer)]
				    [modified-ok (buffer-modified-ok buffer)]
				    [edit (buffer-edit buffer)])
			       (if (and (eq? type 'file)
					(not modified-ok)
					(send edit modified?))
				   (let* ([name (send edit get-filename)]
					  [name (if (null? name)
						    "Untitled"
						    name)])
				     (if (eq? (mred:unsaved-warning name 
								    "Quit")
					      'continue)
					 (begin
					   (set-buffer-modified-ok! buffer #t)
					   (check-buffers))
					 #f))
				   (loop (cdr buffers)))))))]
		    [make-untitled-name
		     (lambda ()
		       (set! untitled-number (add1 untitled-number))
		       (string-append "Untitled " 
				      (number->string untitled-number)))]
		    [clear
		     (lambda ()
		       (if (check-buffers)
			   (begin (set! buffers '())
				  #t)
			   #f))]
		    [pick
		     (lambda (canvas)
		       (let ([which
			      (wx:get-single-choice-index
			       "Select a buffer"
			       "Buffers"
			       (map (lambda (buffer)
				      (let ([name (buffer-name buffer)])
					(if name
					    name
					    (send (buffer-edit buffer) 
						  get-filename))))
				    buffers)
			       () -1 -1 #t 400 200)])
			 (if (>= which 0)
			     (send canvas set-media
				   (buffer-edit (list-ref buffers which))))))])
		   (lambda ()
		     (mred:insert-exit-callback check-buffers))))))
(constant mred:std-buffer-group%)
(define mred:buffer-group% mred:std-buffer-group%)

(define mred:std-frame-group%
  (make-class ()
	      (private
	       [active-frame #f]
	       [frames ()]
	       [frame-counter 0])
	      (class
	       [buffer-group% mred:buffer-group%])
	      (public
	       buffers
	       [get-active-frame
		(lambda ()
		  (cond
		   [active-frame active-frame]
		   [(null? frames) #f]
		   [else (car frames)]))]
	       [set-active-frame
		(lambda (f)
		  (set! active-frame f))]
	       [insert-frame
		(lambda (f)
		  (if (is-a? f mred:std-frame%)
		      (begin 
			(set! frame-counter (add1 frame-counter))
			(send f set-title-prefix
			      (string-append 
			       "MrEd"
			       (if (= 1 frame-counter)
				   ""
				   (string-append
				    " #"
				    (number->string frame-counter)))
			       ": "))
			(set! frames (append frames (list f))))
		      (error "mred:insert-frame: not a standard frame")))]
	       
	       [remove-frame
		(lambda (f)
		  (if (eq? f active-frame)
		      (set! active-frame #f))
		  (catch exit
			  (set! frames
				(let loop ([list frames])
				  (cond
				   [(null? list) ()]
				   [(eq? f (car list)) 
				    (if (and (null? (cdr frames))
					     (not (send buffers 
							check-buffers)))
					(exit #f)
					(cdr list))]
				   [else (cons (car list) 
					       (loop (cdr list)))])))
			  #t))])
	      (lambda ()
		(set! buffers (make-object buffer-group%)))))
(constant mred:std-frame-group%)
(define mred:frame-group% mred:std-frame-group%)

(define mred:frames #f)
(define (mred:keep-frames)
  (set! mred:frames (make-object mred:frame-group%)))

(define mred:unsaved-warning
  (opt-lambda (filename action [can-save-now? #f])
    (let* ([result (void)]
	   [dialog%
	    (make-class wx:dialog-box%
	      (inherit show new-line fit tab)
	      (private
		[on-dont-save
		 (lambda args
		   (set! result 'continue)
		   (show #f))]
		[on-save-now
		 (lambda rags
		   (set! result 'save)
		   (show #f))]
		[on-cancel
		 (lambda args
		   (set! result 'cancel)
		   (show #f))])
	      (lambda ()
		(super-init () "Warning" #t)
		(let ([msg
		       (make-object wx:message% this
			     (string-append "The file \""
					    filename
					    "\" is not saved."))])
		  (new-line)
		  
		  (make-object wx:button% this
			       on-dont-save 
			       (string-append action " anyway"))
		  
		  (tab 50)
		  
		  (let ([now (make-object wx:button% this
					  on-save-now "Save")])
		    (if (not can-save-now?)
			(send now show #f)))
		  
		  (make-object wx:button% this
			       on-cancel "Cancel")
		  (fit)
		  
		  (send msg center wx:const-horizontal))
		
		(show #t)))])
      (make-object dialog%)
      result)))
(constant mred:unsaved-warning)

