/*
 * File:        wx_snip.cc
 * Purpose:     wxSnip implementations
 * Author:      Matthew Flatt
 * Created:     1995
 * Copyright:   (c) 1995, Matthew Flatt
 */

#include "wx.h"
#include "wx_media.h"
#include "wx_ptreq.h"

#define SETWD_NOT_PRESENT 1

#ifdef wx_x
#define CHECK_SPLIT 1
#else
#define CHECK_SPLIT 0
#endif

#ifdef MEMMOVE_NOT_PRESENT
static void memmove(char *dest, char *src, long size)
{
  if (dest < src) {
    while (size--)
      *(dest++) = *(src++);
  } else {
    dest += size;
    src += size;
    while (size--)
      *(--dest) = *(--src);
  }
}
#endif

/* MSW version needs this for just a little while longer... */
#ifndef WXGC_ATOMIC
#define WXGC_ATOMIC
#endif

#define MAX_WASTE 3

#define IMAGE_PIXELS_PER_SCROLL 20
#define IMAGE_VOID_SIZE 20

#define ALWAYSZERO(x) if (x) *x = 0;

/* Text in a text snip can be GC'ed or not. Doesn't matter. */
#define TEXT_USE_MALLOC 1

#if TEXT_USE_MALLOC
#include <stdlib.h>
#define STRALLOC(n) ((char *)malloc(n))
#define STRFREE(s) free(s)
#else
#define STRALLOC(n) (new WXGC_ATOMIC char[n])
#define STRFREE(s) (delete[] s)
#endif

/***************************************************************/

wxSnipClass::wxSnipClass()
{
  classname = "wxbad";
  version = 0;
  required = 0;
  mapPosition = -1;
}

Bool wxSnipClass::ReadHeader(wxMediaStreamIn &)
{
  return TRUE;
}

void wxSnipClass::ReadDone(void)
{
}

Bool wxSnipClass::WriteHeader(wxMediaStreamOut &)
{
  return TRUE;
}

void wxSnipClass::WriteDone(void)
{
}

/***************************************************************/

wxSnip::wxSnip()
{
  count = 1;

  flags = 0;

  snipclass = NULL;

  admin = NULL;

  next = prev = NULL;
  line = NULL;

  WXGC_IGNORE(admin);

  style = wxTheStyleList->BasicStyle();
}

wxSnip::~wxSnip()
{
  next = prev = NULL;
  line = NULL;
}

wxSnip *wxSnip::Next(void)
{
  return next;
}

wxSnip *wxSnip::Previous(void)
{
  return prev;
}

wxSnipAdmin *wxSnip::GetAdmin(void)
{
  return admin;
}

void wxSnip::SetAdmin(wxSnipAdmin *a)
{
  if (PTRNE(a, admin) && (flags & wxSNIP_OWNED))
    return;

  admin = a;
  SizeCacheInvalid();
  if (!a) {
    prev = next = NULL;
    line = NULL;
  } else
    flags |= wxSNIP_OWNED;
}

void wxSnip::SetCount(long new_count)
{
  if (new_count <= 0)
    new_count = 1;
  count = new_count;
  if (admin)
    admin->Resized(this, TRUE);
}

void wxSnip::SetFlags(long new_flags)
{
  /* Make sure that wxSNIP_HARD_NEWLINE implies a wxSNIP_NEWLINE */
  if (new_flags & wxSNIP_HARD_NEWLINE)
    new_flags |= wxSNIP_NEWLINE;
  else if (new_flags & wxSNIP_NEWLINE)
    new_flags -= wxSNIP_NEWLINE;    

  /* Make sure ownership flag doesn't change */
  if (flags & wxSNIP_OWNED)
    new_flags |= wxSNIP_OWNED;
  else if (new_flags & wxSNIP_OWNED)
    new_flags -= wxSNIP_OWNED;
    

  flags = new_flags;
  if (admin)
    admin->Resized(this, TRUE);
}

void wxSnip::OnEvent(wxDC *, float, float, float, float, wxMouseEvent &event)
{
}

void wxSnip::OnChar(wxDC *, float, float, float, float, wxKeyEvent &event)
{
}

void wxSnip::DoEdit(int op, Bool recursive, long time)
{
}

void wxSnip::DoFont(int op, Bool recursive)
{
}

Bool wxSnip::Match(wxSnip *other)
{
  if (PTRNE(other->snipclass, snipclass))
    return FALSE;
  if (other->count != count)
    return FALSE;
  return TRUE;
}
 
void wxSnip::OwnCaret(Bool)
{
}

void wxSnip::SizeCacheInvalid(void)
{
}

void wxSnip::GetExtent(wxDC *, 
		       float, float, 
		       float *w, float *h, 
		       float *descent, float *space,
		       float *lspace, float *rspace)
{
  if (w)
    *w = 0;
  if (h)
    *h = 0;
  if (descent)
    *descent = 0;
  if (space)
    *space = 0;
  if (lspace)
    *lspace = 0;
  if (lspace)
    *lspace = 0;
}

float wxSnip::PartialOffset(wxDC *dc, 
			    float x, float y, long offset)
{
  if (!offset)
    return 0.0;

  float w;
  GetExtent(dc, x, y, &w);
  return w;
}

void wxSnip::Draw(wxDC *dc, float x, float y, 
		  float, float, float, float, 
		  float dx, float dy, Bool)
{
}

void wxSnip::Split(long position, wxSnip **first, wxSnip **second)
{
  wxSnip *snip;

  snip = new wxSnip();
  snip->count = position;
  count -= position;

  *first = snip;
  *second = this;
}

wxSnip *wxSnip::MergeWith(wxSnip *other)
{
    return NULL;
}

void wxSnip::GetText(char *s, long offset, long num)
{
  char *str;

  str = GetText(offset, num, FALSE);
  if (!str)
    memset(s, '.', num);
  else
    memcpy(s, str, num);
}

char *wxSnip::GetText(long offset, long num, Bool flattened)
{
  char *s = new WXGC_ATOMIC char[num + 1];
  memset(s, '.', num);
  s[num] = 0;
  return s;
}

wxSnip *wxSnip::Copy()
{
  wxSnip *snip;

  snip = new wxSnip();
  Copy(snip);

  return snip;
}

void wxSnip::Copy(wxSnip *snip)
{
  snip->count = count;
  snip->flags = flags;
  if (snip->flags & wxSNIP_OWNED)
    snip->flags -= wxSNIP_OWNED;
  snip->snipclass = snipclass;
  snip->style = style;
}

void wxSnip::Write(wxMediaStreamOut &f)
{
}

Bool wxSnip::Resize(float w, float h)
{
  return FALSE;
}

long wxSnip::GetNumScrollSteps()
{
  return 1;
}

long wxSnip::FindScrollStep(float y)
{
  return 0;
}

float wxSnip::GetScrollStepOffset(long i)
{
  return 0;
}

Bool wxSnip::IsOwned(void)
{
  return !!(flags & wxSNIP_OWNED);
}

Bool wxSnip::ReleaseFromOwner(void)
{
  if (!IsOwned())
    return TRUE;

  if (!admin)
    return FALSE;

  if (admin->ReleaseSnip(this)) {
    flags -= wxSNIP_OWNED;
    return TRUE;
  } else
    return FALSE;
}

/***************************************************************/

class TextSnipClass : public wxSnipClass
{
 public:
  TextSnipClass(void);

  virtual wxSnip *Read(wxMediaStreamIn &);
  wxSnip *Read(wxTextSnip *, wxMediaStreamIn &);
};

static TextSnipClass TheTextSnipClass;

TextSnipClass::TextSnipClass(void)
{
  classname = "wxtext";
  version = 1;
  required = TRUE;
}

wxSnip *TextSnipClass::Read(wxMediaStreamIn &f)
{
  return Read(new wxTextSnip(0), f);
}

wxSnip *TextSnipClass::Read(wxTextSnip *snip, wxMediaStreamIn &f)
{
  short flags;
  long count, pos;

  f >> flags;

  pos = f.Tell();
  f >> count;
  f.JumpTo(pos);

  snip->Read(count, f);

  snip->flags = flags;

  return snip;
}

/***************************************************************/

wxTextSnip::wxTextSnip(long allocsize) 
: wxSnip()
{
  flags |= wxSNIP_IS_TEXT | wxSNIP_CAN_APPEND;

  contentsChanged = TRUE;

  allocated = allocsize ? 2 * allocsize : 20;
  text = buffer = STRALLOC(allocated + 1);

  snipclass = &TheTextSnipClass;
  
  count = 0;
}

wxTextSnip::~wxTextSnip()
{
  STRFREE(buffer);
  text = buffer = NULL;
}

void wxTextSnip::SizeCacheInvalid(void)
{
  contentsChanged = TRUE;
}

void wxTextSnip::GetExtent(wxDC *dc, 
			   float x, float y, 
			   float *wo, float *ho, float *dso, float *so,
			   float *ls, float *rs)
{
  char save;

  if (contentsChanged) {
    dc->SetFont(style->GetFont());
    if ((flags & wxSNIP_INVISIBLE) || !count 
	|| (count == 1 && text[0] == '\n')
	|| (count == 1 && text[0] == '\t')) {
      dc->GetTextExtent(" ", &w, &h, &descent, &space);
      if (text[0] != '\t')
	w = 0;
    } else {
      save = text[count];
      text[count] = 0;
      dc->GetTextExtent(text, &w, &h, &descent, &space);
      text[count] = save;
    }
    contentsChanged = FALSE;
  }

  if (wo)
    *wo = w;
  if (ho)
    *ho = h;
  if (dso)
    *dso = descent;
  if (so)
    *so = space;
  if (ls)
    *ls = 0.0;
  if (rs)
    *rs = 0.0;
}

float wxTextSnip::PartialOffset(wxDC *dc, 
				float, float, long offset)
{
  float w, h;
  char save;
  
  if (offset > count)
    offset = count;

  dc->SetFont(style->GetFont());

  save = text[offset];
  text[offset] = 0;
  dc->GetTextExtent(text, &w, &h);
  text[offset] = save;

  return w;
}

void wxTextSnip::Draw(wxDC *dc, float x, float y, 
		      float, float, float, float, 
		      float dx, float dy, Bool)
{
  char save;

  if (flags & wxSNIP_INVISIBLE)
    return;

  save = text[count];
  text[count] = 0;
  dc->DrawText(text, x, y);
  text[count] = save;

  if (style->GetUnderlined()) {
    if (descent >= 2)
      y += h - (descent / 2);
    else
      y += h - descent;
    dc->DrawLine(x, y, x + w, y);
  }
}

void wxTextSnip::Split(long position, wxSnip **first, wxSnip **second)
{
  wxTextSnip *snip;

#if CHECK_SPLIT
  if (position > count) {
    fprintf(stderr, "Bad split %ld of %ld\n", position, count);
    position = count;
  }
#endif

  snip = new wxTextSnip(position);

  contentsChanged = TRUE;

  memcpy(snip->text, text, position);
  text += position;

  snip->count = position;
  count -= position;

  if ((allocated / count) > MAX_WASTE) {
    char *old = buffer;
    allocated = count;
    buffer = STRALLOC(allocated + 1);
    memcpy(buffer, text, count + 1);
    text = buffer;
    STRFREE(old);
  }

  *first = snip;
  *second = this;
}

wxSnip *wxTextSnip::MergeWith(wxSnip *pred)
{
  contentsChanged = TRUE;

  Insert(((wxTextSnip *)pred)->text, pred->count, 0);

  delete pred;
  return this;
}

void wxTextSnip::Insert(char *str, long len, long pos)
{
  if (allocated < count + len) {
    char *oldbuffer = buffer;

    allocated = 2 * (count + len);
    buffer = STRALLOC(allocated + 1);

    memcpy(buffer, text, count);
    STRFREE(oldbuffer);

    text = buffer;
  } else if (PTRNE(text, buffer)
	     && ((text - buffer) + count + len > allocated)) {
    memmove(buffer, text, count);
    text = buffer;
  }
   
  if (pos < count)
    memmove(text + pos + len, text + pos, count - pos);
  memcpy(text + pos, str, len);
  
  count += len;

  contentsChanged = TRUE;
}

void wxTextSnip::GetText(char *s, long offset, long num)
{
  memcpy(s, text + offset, num);
}

char *wxTextSnip::GetText(long offset, long num, Bool flat)
{
  char *s = new WXGC_ATOMIC char[num + 1];
  memcpy(s, text + offset, num);
  s[num] = 0;
  return s;
}

wxSnip *wxTextSnip::Copy()
{
  wxTextSnip *snip = new wxTextSnip(count);

  Copy(snip);

  return snip;
}

void wxTextSnip::Copy(wxTextSnip *snip) 
{
  wxSnip::Copy(snip);
  
  memcpy(snip->text, text, count);
  snip->count = count;

  snip->contentsChanged = TRUE;
}

void wxTextSnip::Write(wxMediaStreamOut &f)
{
  long writeFlags;

  writeFlags = flags;
  if (writeFlags & wxSNIP_OWNED)
    writeFlags -= wxSNIP_OWNED;

  f << writeFlags;
  f.Put(count, text);
}

void wxTextSnip::Read(long len, wxMediaStreamIn &f)
{
  if (allocated < len) {
    allocated = 2 * len;
    STRFREE(buffer);
    buffer = STRALLOC(allocated + 1);
  }

  f.Get((long *)&len, text = buffer);
  count = len;
  contentsChanged = TRUE;
}

/***************************************************************/

class TabSnipClass : public TextSnipClass
{
 public:
  TabSnipClass(void);

  virtual wxSnip *Read(wxMediaStreamIn &);
};

static TabSnipClass TheTabSnipClass;

TabSnipClass::TabSnipClass(void)
{
  classname = "wxtab";
  version = 1;
  required = TRUE;
}

wxSnip *TabSnipClass::Read(wxMediaStreamIn &f)
{
  return TextSnipClass::Read(new wxTabSnip(), f);
}

/***************************************************************/

wxTabSnip::wxTabSnip() : wxTextSnip(1)
{
  snipclass = &TheTabSnipClass;

  flags |= wxSNIP_WIDTH_DEPENDS_ON_X;
  flags -= (flags & wxSNIP_CAN_APPEND);
}

void wxTabSnip::GetExtent(wxDC *dc, 
			  float x, float y, 
			  float *wi, float *h, 
			  float *descent, float *space,
			  float *lspace, float *rspace)
{
  float *tabs, oldw;
  float tabspace;
  int n, i;
  Bool changed;

  changed = contentsChanged;
  oldw = w;

  wxTextSnip::GetExtent(dc, x, y, wi, h, descent, space, lspace, rspace);

  if (changed) {
    /* w is now width of a space */
    float mult;

    if (admin && (admin->GetMedia()->bufferType == wxEDIT_BUFFER)) {
      float space;
      Bool units;

      tabs = ((wxMediaEdit *)admin->GetMedia())->GetTabs(&n, &space, &units);
      tabspace = space;
      mult = units ? 1 : w;
    } else {
      n = 0;
      tabs = NULL;
      tabspace = wxTAB_WIDTH;
      mult = 1;
    }
    
    for (i = 0; i < n; i++)
      if (tabs[i] * mult > x) {
	w = tabs[i] * mult - x;
	break;
      }

    if (i >= n) {
      float base;

      base = tabs ? (tabs[n - 1] * mult) : 0;
      x -= base;

      tabspace *= mult;
      w = base + ((long)tabspace - ((long)x % (long)tabspace));
    }
    
  } else
    w = oldw;

  if (wi)
    *wi = w;
}

float wxTabSnip::PartialOffset(wxDC *dc, float x, float y, long offset)
{
  float w;

  if (!offset)
    return 0;
  else {
    GetExtent(dc, x, y, &w);
    return w;
  }
}

void wxTabSnip::Draw(wxDC *dc, float x, float y, 
		     float, float, float, float, 
		     float dx, float dy, Bool)
{
  /* Do nothing! */
}

wxSnip *wxTabSnip::Copy()
{
  wxTabSnip *snip = new wxTabSnip();

  wxTextSnip::Copy(snip);

  return snip;
}

/***************************************************************/

class ImageSnipClass : public wxSnipClass
{
 public:
  ImageSnipClass(void);

  virtual wxSnip *Read(wxMediaStreamIn &);
};

static ImageSnipClass TheImageSnipClass;

ImageSnipClass::ImageSnipClass(void)
{
  classname = "wximage";
  version = 1;
  required = FALSE;
}

wxSnip *ImageSnipClass::Read(wxMediaStreamIn &f)
{
  wxImageSnip *snip;
  char *filename;
  long type;
  Bool relative;
  float w, h, dx, dy;

  filename = f.GetString(NULL);
  f >> type;
  f >> w;
  f >> h;
  f >> dx;
  f >> dy;
  f >> relative;

  snip = new wxImageSnip(filename, type, relative);

  delete[] filename;

  snip->Resize(w, h);
  snip->SetOffset(dx, dy);

  return (wxSnip *)snip;
}

/***************************************************************/

wxImageSnip::wxImageSnip(char *name, long type, Bool relative) : wxSnip()
{
  snipclass = &TheImageSnipClass;

  contentsChanged = TRUE;
  filename = NULL;
  filetype = 0;
  relativePath = FALSE;
  fromfile = TRUE;
  bm = NULL;
  vieww = viewh = -1.0;
  viewdx = viewdy = 0.0;
  offscreen = new wxMemoryDC();

  if (name)
    LoadFile(name, type, relative);
}

wxImageSnip::~wxImageSnip()
{
  if (filename)
    delete[] filename;

  offscreen->SelectObject(NULL);
  delete offscreen;

  if (bm && fromfile)
    delete bm;
}

void wxImageSnip::SizeCacheInvalid(void)
{
  contentsChanged = TRUE;  
}


void wxImageSnip::GetExtent(wxDC *dc,
			    float x, float y,
			    float *wi, float *hi, 
			    float *descent, float *space,
			    float *lspace, float *rspace)
{
  if (contentsChanged) {
    if (filename && bm && bm->Ok()) {
      if (viewh <= 0)
	h = bm->GetHeight();
      else
	h = viewh;
      if (vieww <= 0)
	w = bm->GetWidth();
      else
	w = vieww;
    } else {
      h = w = 0;
    }
    if (!h)
      h = IMAGE_VOID_SIZE;
    if (!w)
      w = IMAGE_VOID_SIZE;
  }

  if (wi)
    *wi = w;
  if (hi)
    *hi = h;
  ALWAYSZERO(descent);
  ALWAYSZERO(space);
  ALWAYSZERO(lspace);
  ALWAYSZERO(rspace);
}

void wxImageSnip::Draw(wxDC *dc, float x, float y, 
		       float l, float t, float r, float b,
		       float dx, float dy, Bool)
{
  float lx, ly, lt, ll, lw, lh;

  if ((fromfile && !filename) || !bm->Ok()) {
    dc->DrawRectangle(x + 1, y + 1, w - 2, h - 2);
    dc->DrawLine(x + 1, y + 1, x + w - 1, y + h - 1);
    dc->DrawLine(x + 1, y + h - 2, x + w - 1, y);
    return;
  }

  if (vieww > 0 && r > x + vieww)
    r = x + vieww;
  if (x + w < r)
    r = w + x;
  if (viewh > 0 && b > y + viewh)
    b = y + viewh;
  if (y + h < b)
    b = y + h;

  x += viewdx;
  y += viewdy;

  if ((l > x + w) || (t > y + h))
    return;
  
  lx = l - x;
  if (lx < 0) {
    lx = 0;
    ll = x;
  } else
    ll = l;
    
  ly = t - y;
  if (ly < 0) {
    ly = 0;
    lt = y;
  } else
    lt = t;

  lw = r - ll;
  lh = b - lt;

  dc->Blit(ll, lt, lw, lh, offscreen, lx, ly, wxCOPY);
}

wxSnip *wxImageSnip::Copy(void)
{
  wxImageSnip *snip;

  snip = new wxImageSnip();
  Copy(snip);

  return (wxSnip *)snip;
}

void wxImageSnip::Write(wxMediaStreamOut &f)
{
  f << ((fromfile && filename) ? filename : "");
  f << filetype;
  f << vieww;
  f << viewh;
  f << viewdx;
  f << viewdy;
  f << relativePath;
}

void wxImageSnip::LoadFile(char *name, long type, Bool relative)
{
  if (filename)
    delete[] filename;
 
  if (name && !*name)
    name = NULL;
   
  if (relative && name) {
    if (name[0] == '/')
      relative = FALSE;
#ifdef wx_msw
    if (name[0] == '\\')
      relative = FALSE;
    if (name[0] && name[1] == ':')
      relative = FALSE;
#endif
#ifdef wx_x
    if (name[0] == '~')
      relative = FALSE;
#endif
  }

  relativePath = relative && name;

  if (relativePath)
    flags |= wxSNIP_USES_BUFFER_PATH;
  else if (flags & wxSNIP_USES_BUFFER_PATH)
    flags -= wxSNIP_USES_BUFFER_PATH;

  if (name) {
    char *loadname;

    loadname = name;

    if (!relativePath || admin) {
      if (relativePath) {
	wxMediaBuffer *b;
	char *fn, *path;
	
	b = admin ? admin->GetMedia() : NULL;
	fn = b ? b->GetFilename() : NULL;
	if (fn) {
	  path = wxPathOnly(fn);
	  if (path) {
	    loadname = new WXGC_ATOMIC char[strlen(path) + strlen(name) + 2];
	    strcpy(loadname, path);
#ifdef wx_x
	    strcat(loadname, "/");
#else
	    strcat(loadname, "\\");
#endif	    
	    strcat(loadname, name);
	  }
	}
      }
      
      if (!bm || !fromfile)
	bm = new wxBitmap(loadname, type);
      else
	bm->LoadFile(loadname, type);
      
      if (bm->Ok())
	offscreen->SelectObject(bm);
    }

    filename = copystring(name);
    filetype = type;
    fromfile = TRUE;
  } else {
    filename = NULL;
  }

  contentsChanged = TRUE;

  if (admin)
    admin->Resized(this, TRUE);
}

void wxImageSnip::Copy(wxImageSnip *newSnip)
{
  wxSnip::Copy(newSnip);
  if (fromfile)
    newSnip->LoadFile(filename, filetype, relativePath);
  else
    newSnip->SetBitmap(bm);
}

char *wxImageSnip::GetFilename(Bool *rel)
{
  if (rel)
    *rel = fromfile && filename && relativePath;

  return !fromfile ? NULL : filename;
}

long wxImageSnip::GetFiletype()
{
  return !fromfile ? 0 : filetype;
}

void wxImageSnip::SetBitmap(wxBitmap *map)
{
  if (bm) {
    offscreen->SelectObject(NULL);
    if (fromfile)
      delete bm;
  }

  bm = map;
  offscreen->SelectObject(map);
  fromfile = FALSE;

  contentsChanged = TRUE;

  if (admin)
    admin->Resized(this, TRUE);
}

void wxImageSnip::SetOffset(float x, float y)
{
  viewdx = x;
  viewdy = y;

  contentsChanged = TRUE;

  if (admin)
    admin->NeedsUpdate(this, 0, 0, w, h);
}

Bool wxImageSnip::Resize(float w, float h)
{
  if (w <= 0 && !fromfile)
    w = IMAGE_VOID_SIZE;
  if (h <= 0 && !fromfile)
    h = IMAGE_VOID_SIZE;

  vieww = w;
  viewh = h;

  contentsChanged = TRUE;

  if (admin)
    admin->Resized(this, TRUE);

  return TRUE;
}

long wxImageSnip::GetNumScrollSteps()
{
  long ss;

  ss = (long)(h / IMAGE_PIXELS_PER_SCROLL);

  return ss ? ss : 1;
}

long wxImageSnip::FindScrollStep(float y)
{
  return (long)(y / IMAGE_PIXELS_PER_SCROLL);
}

float wxImageSnip::GetScrollStepOffset(long i)
{
  return ((float )i) * IMAGE_PIXELS_PER_SCROLL;
}

void wxImageSnip::SetAdmin(wxSnipAdmin *a)
{
  if (PTRNE(admin, a))
    wxSnip::SetAdmin(a);
  if (admin && relativePath && fromfile && filename)
    LoadFile(filename, filetype, TRUE);
}

/***************************************************************/

class MediaSnipClass : public wxSnipClass
{
 public:
  MediaSnipClass(void);

  virtual wxSnip *Read(wxMediaStreamIn &);
};

static MediaSnipClass TheMediaSnipClass;

MediaSnipClass::MediaSnipClass(void)
{
  classname = "wxmedia";
  version = 1;
  required = TRUE;
}

wxSnip *MediaSnipClass::Read(wxMediaStreamIn &f)
{
  wxMediaBuffer *media;
  wxMediaSnip *snip;
  Bool border;
  int lm, tm, rm, bm, li, ti, ri, bi, type;
  float w, W, h, H;

  f >> type >> border >> lm >> tm >> rm >> bm >> li >> ti >> ri >> bi
    >> w >> W >> h >> H;

  if (type == wxEDIT_BUFFER)
    media = new wxMediaEdit();
  else
    media = new wxMediaPasteboard();

  snip = new wxMediaSnip(media, border, lm, tm, rm, bm, li, ti, ri, bi,
			 w, W, h, H);

  media->ReadFromFile(f);

  return snip;
}

/***************************************************************/

wxSnipClassList::wxSnipClassList(void)
: wxList((int)wxKEY_STRING)
{
}

wxSnipClassList::~wxSnipClassList()
{
}

wxSnipClass *wxSnipClassList::Find(char *name)
{
  wxNode *node;

  node = wxList::Find(name);

  return node ? (wxSnipClass *)node->Data() : NULL;
}

short wxSnipClassList::FindPosition(wxSnipClass *sclass)
{
  wxNode *node;
  short i;
  
  for (i = 0, node = First(); node; node = node->Next(), i++)
    if (PTREQ(sclass, (wxSnipClass *)node->Data()))
      return i;

  return -1;
}

void wxSnipClassList::Add(wxSnipClass *snipclass)
{
  if (wxList::Find(snipclass->classname))
    return;

  Append(snipclass->classname, snipclass);
}

int wxSnipClassList::Number(void)
{
  return wxList::Number();
}

wxSnipClass *wxSnipClassList::Nth(int n)
{
  return (wxSnipClass *)wxList::Nth(n)->Data();
}

/***************************************************************/

wxStandardSnipClassList::wxStandardSnipClassList(void)
{
  Add(&TheTextSnipClass);
  Add(&TheTabSnipClass);
  Add(&TheMediaSnipClass);
  Add(&TheImageSnipClass);
}

void wxStandardSnipClassList::ResetHeaderFlags(int doneMsg)
{
  wxNode *node;
  wxSnipClass *sclass;

  for (node = First(); node; node = node->Next()) {
    sclass = (wxSnipClass *)node->Data();
    if (sclass->headerFlag) {
      if (doneMsg == wxRESET_DONE_WRITE)
	sclass->WriteDone();
      else if (doneMsg == wxRESET_DONE_READ)
	sclass->ReadDone();
    }
    sclass->headerFlag = 0;
    sclass->mapPosition = -1;
  }
}

Bool wxStandardSnipClassList::Write(wxMediaStreamOut &f)
{
  wxNode *node;
  wxSnipClass *sclass;
  short i;

  f << Number();

  for (i = 0, node = First(); node; node = node->Next(), i++) {
    sclass = (wxSnipClass *)node->Data();
    f << sclass->classname;
    f << sclass->version;
    f << sclass->required;
    sclass->mapPosition = i;
  }

  return TRUE;
}

Bool wxStandardSnipClassList::Read(wxMediaStreamIn &f)
{
  int count, i;
  long n;
  wxSnipClass *sclass;
  char buffer[256];
  int version;
  Bool required;

  f >> count;

  buffer[255] = 0;

  for (i = 0; i < count; i++) {
    n = 255;
    f.Get((long *)&n, (char *)buffer);
    f >> version;
    f >> required;
    if (!f.Ok())
      return FALSE;
    sclass = Find(buffer);
    if (!sclass) {
      char buffer2[256];
      sprintf(buffer2, "Unknown snip class \"%.100s\".", buffer);
      wxMessageBox(buffer2, "Warning");
    } else
      sclass->mapPosition = FindPosition(sclass);
  }

  return TRUE;
}

wxSnipClass *wxStandardSnipClassList::FindByMapPosition(short n)
{
  wxNode *node;
  wxSnipClass *sclass;
  
  if (n < 0)
    return NULL;

  for (node = First(); node; node = node->Next()) {
    sclass = (wxSnipClass *)node->Data();
    if (sclass->mapPosition == n)
      return sclass;
  }

  return NULL;
}

wxStandardSnipClassList wxTheSnipClassList;

/***************************************************************/

wxBufferDataClass::wxBufferDataClass()
{
  classname = "wxbad";
  required = 0;
}

wxBufferData::wxBufferData()
{
  next = NULL;
}

wxBufferData::~wxBufferData()
{
  if (next)
    delete next;
}

class LocationBufferDataClass : public wxBufferDataClass
{
 public:
  LocationBufferDataClass();
  wxBufferData *Read(wxMediaStreamIn &);
};

LocationBufferDataClass::LocationBufferDataClass()
{
  classname = "wxloc";
  required = 1;
}

wxBufferData *LocationBufferDataClass::Read(wxMediaStreamIn &f)
{
  wxLocationBufferData *data;

  data = new wxLocationBufferData;
  f >> data->x;
  f >> data->y;

  return data;
}

LocationBufferDataClass TheLocationBufferDataClass;

wxLocationBufferData::wxLocationBufferData()
{
  x = y = 0;
  dataclass = &TheLocationBufferDataClass;
}

Bool wxLocationBufferData::Write(wxMediaStreamOut &f)
{
  f << x;
  f << y;
  return TRUE;
}

/**************************************************************/

wxBufferDataClassList::wxBufferDataClassList(void)
: wxList((int)wxKEY_STRING)
{
  Add(&TheLocationBufferDataClass);
}

wxBufferDataClassList::~wxBufferDataClassList()
{
}

wxBufferDataClass *wxBufferDataClassList::Find(char *name)
{
  wxNode *node;

  node = wxList::Find(name);

  return node ? (wxBufferDataClass *)node->Data() : NULL;
}

short wxBufferDataClassList::FindPosition(wxBufferDataClass *sclass)
{
  wxNode *node;
  short i;
  
  for (i = 0, node = First(); node; node = node->Next(), i++)
    if (PTREQ(sclass, (wxBufferDataClass *)node->Data()))
      return i + 1;

  return 0;
}

void wxBufferDataClassList::Add(wxBufferDataClass *dataclass)
{
  if (wxList::Find(dataclass->classname))
    return;

  Append(dataclass->classname, dataclass);
}

int wxBufferDataClassList::Number(void)
{
  return wxList::Number();
}

wxBufferDataClass *wxBufferDataClassList::Nth(int n)
{
  return (wxBufferDataClass *)wxList::Nth(n)->Data();
}

Bool wxBufferDataClassList::Write(wxMediaStreamOut &f)
{
  wxNode *node;
  wxBufferDataClass *sclass;
  short i;

  f << Number();

  for (i = 0, node = First(); node; node = node->Next(), i++) {
    sclass = (wxBufferDataClass *)node->Data();
    f << sclass->classname;
    sclass->mapPosition = i + 1;
  }

  return TRUE;
}

Bool wxBufferDataClassList::Read(wxMediaStreamIn &f)
{
  int count, i;
  long n;
  wxBufferDataClass *sclass;
  char buffer[256];

  f >> count;

  buffer[255] = 0;

  for (i = 0; i < count; i++) {
    n = 255;
    f.Get((long *)&n, (char *)buffer);
    if (!f.Ok())
      return FALSE;
    sclass = Find(buffer);
    if (!sclass) {
      char buffer2[256];
      sprintf(buffer2, "Unknown snip data class \"%.100s\".", buffer);
      wxMessageBox(buffer2, "Warning");
    } else
      sclass->mapPosition = FindPosition(sclass);
  }

  return TRUE;
}

wxBufferDataClass *wxBufferDataClassList::FindByMapPosition(short n)
{
  wxNode *node;
  wxBufferDataClass *sclass;
  
  if (n <= 0)
    return NULL;

  for (node = First(); node; node = node->Next()) {
    sclass = (wxBufferDataClass *)node->Data();
    if (sclass->mapPosition == n)
      return sclass;
  }

  return NULL;
}

wxBufferDataClassList wxTheBufferDataClassList;
